from collections import namedtuple

Command = namedtuple('Command', 'name impl')


def command():
    def decorator(func):
        func._is_cmd = True
        return func

    return decorator


class CommandsMeta(type):
    def __new__(cls, name, bases, dct):
        commands = dct.get('commands', [])

        for member_name, member_value in dct.items():
            if getattr(member_value, '_is_cmd', False):
                commands.append(Command(name=member_name,
                                        impl=member_value))

        dct['commands'] = commands
        return super(CommandsMeta, cls).__new__(cls, name, bases, dct)



import lldb
import pytest

from hilda.hilda_client import HildaClient


@pytest.fixture(scope='function')
def hilda_client(lldb_debugger):
    client = HildaClient(lldb_debugger)
    client.init_dynamic_environment()
    lldb.hilda_client = client
    with client.stopped():
        yield client


@pytest.fixture(scope='session', autouse=True)
def disable_jetsam_memory_checks(lldb_debugger):
    client = HildaClient(lldb_debugger)
    lldb.hilda_client = client
    client.init_dynamic_environment()
    client.disable_jetsam_memory_checks()



import lldb


def _CFRunLoopServiceMachPort_hook(hilda, *args):
    """
    :param hilda.hilda_client.HildaClient hilda:
    """
    hilda.jump(hilda.CFRunLoopServiceMachPort_while_ea)
    hilda.cont()


def disable_mach_msg_errors():
    """
    Disable the error check inside the CFRunLoopServiceMachPort from the mach_msg syscall.
    This is used to debug slow handling of mach messages.
    """
    hilda = lldb.hilda_client
    with hilda.stopped():
        instructions = hilda.symbols.__CFRunLoopServiceMachPort.disass(2000, should_print=False)
        while_ea = None
        for instruction in instructions:
            if (while_ea is None) and instruction.DoesBranch():
                # Beginning of the `while(true) { ... }`
                while_ea = instruction.GetOperands(hilda.target)
                hilda.CFRunLoopServiceMachPort_while_ea = int(hilda.file_symbol(eval(while_ea)))
            elif instruction.GetMnemonic(hilda.target) in ('brk', 'ud2'):
                symbol = hilda.symbol(instruction.addr.GetLoadAddress(hilda.target))
                symbol.bp(
                    _CFRunLoopServiceMachPort_hook,
                    forced=True,
                    name=f'__CFRunLoopServiceMachPort-brk-{int(symbol - hilda.symbols.__CFRunLoopServiceMachPort)}'
                )

        # on iOS 16.x, will need to also patch this one
        handle_error = hilda.symbols.get('__CFRunLoopServiceMachPort.cold.1')
        if handle_error is None:
            return

        for instruction in handle_error.disass(2000, should_print=False):
            if instruction.GetMnemonic(hilda.target) in ('brk', 'ud2'):
                # mov x0, x0
                hilda.symbol(instruction.addr.GetLoadAddress(hilda.target)).poke(b'\xe0\x03\x00\xaa')



import json
import os
from contextlib import suppress
from dataclasses import dataclass
from functools import partial
from pathlib import Path

from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer
from objc_types_decoder.decode import decode as decode_type

from hilda.exceptions import HildaException
from hilda.objective_c_class import Class, convert_encoded_property_attributes, Method, Property
from hilda.symbols_jar import SymbolsJar
from hilda.symbol import Symbol


class SettingIvarError(HildaException):
    """ Raise when trying to set an Ivar too early or when the Ivar doesn't exist. """
    pass


@dataclass
class Ivar:
    name: str
    value: Symbol
    type_: str
    offset: int


class ObjectiveCSymbol(Symbol):
    """
    Wrapper object for an objective-c symbol.
    Allowing easier access to its properties, methods and ivars.
    """

    @classmethod
    def create(cls, value: int, client):
        """
        Create an ObjectiveCSymbol object.
        :param value: Symbol address.
        :param hilda.hilda_client.HildaClient client: hilda client.
        :return: ObjectiveCSymbol object.
        :rtype: ObjectiveCSymbol
        """
        symbol = super(ObjectiveCSymbol, cls).create(value, client)
        symbol.ivars = []
        symbol.properties = []
        symbol.methods = []
        symbol.class_ = None  # type: Class
        symbol.reload()
        return symbol

    def reload(self):
        """
        Reload object's in-memory layout.
        """
        self.ivars.clear()
        self.properties.clear()
        self.methods.clear()
        self.class_ = None

        with open(os.path.join(Path(__file__).resolve().parent, 'get_objectivec_symbol_data.m'), 'r') as f:
            obj_c_code = f.read()

        obj_c_code = obj_c_code.replace('__symbol_address__', f'{self:d}')
        data = json.loads(self._client.po(obj_c_code))

        self._reload_ivars(data['ivars'])
        self._reload_properties(data['properties'])
        self.methods = [Method.from_data(method, self._client) for method in data['methods']]

        data['name'] = data['class_name']
        data['address'] = data['class_address']
        data['super'] = data['class_super']
        self.class_ = Class(self._client, self._client.symbol(data['class_address']), data)

    def show(self, recursive: bool = False):
        """
        Print to terminal the highlighted class description.
        :param recursive: Show methods of super classes.
        """
        print(highlight(self._to_str(recursive), ObjectiveCLexer(), TerminalTrueColorFormatter(style='native')))

    def objc_call(self, selector: str, *params):
        """
        Make objc_call() from self return ObjectiveCSymbol when it's an objc symbol.
        :param selector: Selector to execute.
        :param params: Additional parameters.
        :return: ObjectiveCSymbol when return type is an objc symbol.
        """
        symbol = super(ObjectiveCSymbol, self).objc_call(selector, *params)
        return symbol.objc_symbol if self._client.is_objc_type(symbol) else symbol

    def _reload_ivars(self, ivars_data):
        raw_ivars = sorted(ivars_data, key=lambda ivar: ivar['offset'])
        for i, ivar in enumerate(raw_ivars):
            ivar_type = decode_type(ivar['type'])
            value = ivar['value']
            if i < len(raw_ivars) - 1:
                # The .fm file returns a 64bit value, regardless of the real size.
                size = raw_ivars[i + 1]['offset'] - ivar['offset']
                value = value & ((2 ** (size * 8)) - 1)
            ivar_value = self._client.symbol(value)
            self.ivars.append(Ivar(name=ivar['name'], type_=ivar_type, offset=ivar['offset'], value=ivar_value))

    def _reload_properties(self, properties_data):
        for prop in properties_data:
            prop_attributes = convert_encoded_property_attributes(prop['attributes'])
            self.properties.append(Property(name=prop['name'], attributes=prop_attributes))

    def _set_ivar(self, name, value):
        try:
            ivars = self.__getattribute__('ivars')
            class_name = self.__getattribute__('class_').name
        except AttributeError as e:
            raise SettingIvarError from e

        for i, ivar in enumerate(ivars):
            if ivar.name == name:
                size = self.item_size
                if i < len(self.ivars) - 1:
                    size = ivars[i + 1].offset - ivar.offset
                with self.change_item_size(size):
                    self[ivar.offset // size] = value
                    ivar.value = value
                return
        raise SettingIvarError(f'Ivar "{name}" does not exist in "{class_name}"')

    def _to_str(self, recursive=False):
        protocols_buf = f'<{",".join(self.class_.protocols)}>' if self.class_.protocols else ''

        if self.class_.super is not None:
            buf = f'@interface {self.class_.name}: {self.class_.super.name} {protocols_buf}\n'
        else:
            buf = f'@interface {self.class_.name} {protocols_buf}\n'

        # Add ivars
        buf += '{\n'
        for ivar in self.ivars:
            buf += f'\t{ivar.type_} {ivar.name} = 0x{int(ivar.value):x}; // 0x{ivar.offset:x}\n'
        buf += '}\n'

        # Add properties
        for prop in self.properties:
            attrs = prop.attributes
            buf += f'@property ({",".join(attrs.list)}) {prop.attributes.type_} {prop.name};\n'

            if attrs.synthesize is not None:
                buf += f'@synthesize {prop.name} = {attrs.synthesize};\n'

        # Add methods
        methods = self.methods.copy()

        # Add super methods.
        if recursive:
            for sup in self.class_.iter_supers():
                for method in filter(lambda m: m not in methods, sup.methods):
                    methods.append(method)

        # Print class methods first.
        methods.sort(key=lambda m: not m.is_class)

        for method in methods:
            buf += str(method)

        buf += '@end'
        return buf

    @property
    def symbols_jar(self) -> SymbolsJar:
        """ Get a SymbolsJar object for quick operations on all methods """
        jar = SymbolsJar.create(self._client)

        for m in self.methods:
            jar[m.name] = m.address

        return jar

    def __dir__(self):
        result = set()

        for ivar in self.ivars:
            result.add(ivar.name)

        for method in self.methods:
            result.add(method.name.replace(':', '_'))

        for sup in self.class_.iter_supers():
            for method in sup.methods:
                result.add(method.name.replace(':', '_'))

        result.update(list(super(ObjectiveCSymbol, self).__dir__()))
        return list(result)

    def __getitem__(self, item):
        if isinstance(item, int):
            return super(ObjectiveCSymbol, self).__getitem__(item)

        # Ivars
        for ivar in self.ivars:
            if ivar.name == item:
                if self._client.is_objc_type(ivar.value):
                    return ivar.value.objc_symbol
                return ivar.value

        # Properties
        for prop in self.properties:
            if prop.name == item:
                return self.objc_call(item)

        # Methods
        for method in self.methods:
            if method.name == item:
                return partial(self.class_.objc_call, item) if method.is_class else partial(self.objc_call, item)

        for sup in self.class_.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    return partial(self.class_.objc_call, item) if method.is_class else partial(self.objc_call, item)

        raise AttributeError(f''''{self.class_.name}' has no attribute {item}''')

    def __getattr__(self, item: str):
        return self[self.class_.sanitize_name(item)]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            super(ObjectiveCSymbol, self).__setitem__(key, value)
            return

        with suppress(SettingIvarError):
            self._set_ivar(key, value)
            return

    def __setattr__(self, key, value):
        try:
            key = self.__getattribute__('class_').sanitize_name(key)
        except AttributeError:
            pass
        try:
            self._set_ivar(key, value)
        except SettingIvarError:
            super(ObjectiveCSymbol, self).__setattr__(key, value)

    def __str__(self):
        return self._to_str(False)



from cached_property import cached_property

import lldb

from hilda.snippets.macho.all_image_infos import AllImageInfos
from hilda.snippets.syslog import open_syslog_socket


def all_image_infos():
    return AllImageInfos()


@cached_property
def version():
    with lldb.hilda_client.stopped(1):
        return \
            lldb.hilda_client.symbols.dyldVersionString.peek_str().decode('utf-8').split("PROJECT", 1)[1].split("\n")[0]


def enable_syslog():
    client = lldb.hilda_client
    with client.stopped(1):
        client.symbols._ZZN4dyldL9useSyslogEvE14launchdChecked.item_size = 1
        client.symbols._ZZN4dyldL9useSyslogEvE14launchdChecked[0] = 1
        client.symbols._ZZN4dyldL9useSyslogEvE12launchdOwned.item_size = 1
        client.symbols._ZZN4dyldL9useSyslogEvE12launchdOwned[0] = 1
        client.symbols._ZN4dyldL10sLogSocketE.item_size = 4
        client.symbols._ZN4dyldL10sLogSocketE[0] = 0xFFFFFFFF
        options = (
            'DYLD_PRINT_APIS',
            'DYLD_PRINT_APIS_APP',
            'DYLD_PRINT_BINDINGS'
            'DYLD_PRINT_DOFS',
            'DYLD_PRINT_INITIALIZERS',
            'DYLD_PRINT_INTERPOSING',
            'DYLD_PRINT_LIBRARIES',
            'DYLD_PRINT_LIBRARIES_POST_LAUNCH',
            'DYLD_PRINT_NOTIFICATIONS',
            'DYLD_PRINT_STATSTICS'
            'DYLD_PRINT_STATSTICS_DETAILS',
            'DYLD_PRINT_SEGMENTS',
            'DYLD_PRINT_WEAK_BINDINGS',
            'DYLD_PRINT_OPTS',
            'DYLD_PRINT_WARNINGS',
        )

        open_syslog_socket()

        with client.safe_malloc(8 * len(options) + 1) as envp:
            for i, option in enumerate(options):
                # enable for /usr/lib/dyld via dyld::processDyldEnvironmentVariable(option)
                client.symbols._ZN4dyld30processDyldEnvironmentVariableEPKcS1_S1_(option)
                variable = f'{option}=1'
                envp[i] = client.symbols.malloc(len(option) + 10)
                envp[i].poke(variable)

            envp[len(options)] = 0

            # enable for libdyld.dylib via dyld::setLoggingFromEnvs(eng)
            client.symbols._ZN5dyld318setLoggingFromEnvsEPPKc(envp)



import json

import lldb


def dirlist(path: str):
    hilda = lldb.hilda_client
    return json.loads(hilda.po(f'''
    NSArray *objectData = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:@"{path}" error:Nil];
    NSData *data = [NSJSONSerialization dataWithJSONObject:objectData options:0 error:nil];
    [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
    '''))



from contextlib import suppress

import lldb

from hilda.exceptions import SymbolAbsentError, AddingLldbSymbolError


class SymbolsJar(dict):
    @staticmethod
    def create(client):
        """
        Factory method for creating symbols jars
        :param client: Hilda client
        :rtype: SymbolsJar
        """
        jar = SymbolsJar()
        jar.__dict__['_client'] = client
        return jar

    def get_lazy(self, name):
        client = self.__dict__['_client']
        if '{' in name:
            # remove module name from symbol
            name = name.split('{', 1)[0]
        for s in client.target.FindSymbols(name):
            with suppress(AddingLldbSymbolError):
                return client.add_lldb_symbol(s.symbol)
        return None

    def __getitem__(self, item):
        if item not in self:
            symbol = self.get_lazy(item)
            if symbol:
                return symbol
        return dict.__getitem__(self, item)

    def __getattr__(self, name):
        if name not in self:
            client = self.__dict__['_client']
            for s in client.target.FindSymbols(name):
                with suppress(AddingLldbSymbolError):
                    return client.add_lldb_symbol(s.symbol)
            raise SymbolAbsentError(f'no such symbol: {name}')

        return self.get(name)

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)

    def __delattr__(self, item):
        return self.__delitem__(item)

    def __sub__(self, other):
        retval = SymbolsJar.create(self.__dict__['_client'])
        for k1, v1 in self.items():
            if k1 not in other:
                retval[k1] = v1
        return retval

    def __add__(self, other):
        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in other.items():
            retval[k] = v
        for k, v in self.items():
            retval[k] = v
        return retval

    def bp(self, callback=None, **args):
        """
        Place a breakpoint on all symbols in current jar.
        Look for the bp command for more details.
        :param callback:  callback function to be executed upon an hit
        :param args: optional args for the bp command
        """
        for k, v in self.items():
            v.bp(callback, **args)

    def by_module(self):
        """
        Filter to only names containing their module names
        :return: reduced symbol jar
        """
        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in self.items():
            if '{' not in k or '}' not in k:
                continue
            retval[k] = v
        return retval

    def by_type(self, lldb_type):
        """
        Filter by LLDB symbol types (for example: lldb.eSymbolTypeCode,
        lldb.eSymbolTypeData, ...)
        :param lldb_type: symbol type from LLDB consts
        :return: symbols matching the type filter
        """
        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in self.items():
            if v.type_ == lldb_type:
                retval[k] = v
        return retval

    def code(self):
        """
        Filter only code symbols
        :return: symbols with type lldb.eSymbolTypeCode
        """
        return self.by_type(lldb.eSymbolTypeCode)

    def data(self):
        """
        Filter only data symbols
        :return: symbols with type lldb.eSymbolTypeCode
        """
        return self.by_type(lldb.eSymbolTypeData)

    def objc_class(self):
        """
        Filter only objc meta classes
        :return: symbols with type lldb.eSymbolTypeObjCMetaClass
        """
        return self.clean().by_type(lldb.eSymbolTypeObjCMetaClass)

    def clean(self):
        """
        Filter only symbols without module suffix
        :return: reduced symbol jar
        """
        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in self.items():
            if '{' in k:
                continue
            retval[k] = v
        return retval

    def monitor(self, **args):
        """
        Perform monitor for all symbols in current jar.
        See monitor command for more details.
        :param args: given arguments for monitor command
        """
        for name, address in self.items():
            if name == '_client':
                continue
            name = args.get('name', name)
            address.monitor(name=name, **args)

    def startswith(self, exp, case_sensitive=True):
        """
        Filter only symbols with given prefix
        :param exp: prefix
        :param case_sensitive: is case sensitive
        :return: reduced symbol jar
        """
        if not case_sensitive:
            exp = exp.lower()

        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in self.items():
            orig_k = k
            if not case_sensitive:
                k = k.lower()
            if k.startswith(exp):
                retval[orig_k] = v
        return retval

    def endswith(self, exp, case_sensitive=True):
        """
        Filter only symbols with given prefix
        :param exp: prefix
        :param case_sensitive: is case sensitive
        :return: reduced symbol jar
        """
        if not case_sensitive:
            exp = exp.lower()

        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in self.items():
            orig_k = k
            if not case_sensitive:
                k = k.lower()
            if k.endswith(exp):
                retval[orig_k] = v
        return retval

    def find(self, exp, case_sensitive=True):
        """
        Filter symbols containing a given expression
        :param exp: given expression
        :param case_sensitive: is case sensitive
        :return: reduced symbol jar
        """
        if not case_sensitive:
            exp = exp.lower()

        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in self.items():
            orig_k = k
            if not case_sensitive:
                k = k.lower()
            if exp in k:
                retval[orig_k] = v
        return retval



import lldb

from construct import Hex, Int32ul, Pointer, PaddedString, Int64ul, this, Enum, Struct, Tell, Switch, Bytes, Seek, \
    Array, Int8ul

from hilda.snippets.macho.apple_version import version_t
from hilda.symbol import SymbolFormatField

# See: https://opensource.apple.com/source/xnu/xnu-7195.81.3/EXTERNAL_HEADERS/mach-o/loader.h
LC_REQ_DYLD = 0x80000000

LOAD_COMMAND_TYPE = Enum(Int32ul,
                         LC_SEGMENT=0x1,
                         LC_SYMTAB=0x2,
                         LC_SYMSEG=0x3,
                         LC_THREAD=0x4,
                         LC_UNIXTHREAD=0x5,
                         LC_LOADFVMLIB=0x6,
                         LC_IDFVMLIB=0x7,
                         LC_IDENT=0x8,
                         LC_FVMFILE=0x9,
                         LC_PREPAGE=0xa,
                         LC_DYSYMTAB=0xb,
                         LC_LOAD_DYLIB=0xc,
                         LC_ID_DYLIB=0xd,
                         LC_LOAD_DYLINKER=0xe,
                         LC_ID_DYLINKER=0xf,
                         LC_PREBOUND_DYLIB=0x10,
                         LC_ROUTINES=0x11,
                         LC_SUB_FRAMEWORK=0x12,
                         LC_SUB_UMBRELLA=0x13,
                         LC_SUB_CLIENT=0x14,
                         LC_SUB_LIBRARY=0x15,
                         LC_TWOLEVEL_HINTS=0x16,
                         LC_PREBIND_CKSUM=0x17,
                         LC_LOAD_WEAK_DYLIB=(0x18 | LC_REQ_DYLD),
                         LC_SEGMENT_64=0x19,
                         LC_ROUTINES_64=0x1a,
                         LC_UUID=0x1b,
                         LC_RPATH=(0x1c | LC_REQ_DYLD),
                         LC_CODE_SIGNATURE=0x1d,
                         LC_SEGMENT_SPLIT_INFO=0x1e,
                         LC_REEXPORT_DYLIB=0x1f | LC_REQ_DYLD,
                         LC_LAZY_LOAD_DYLIB=0x20,
                         LC_ENCRYPTION_INFO=0x21,
                         LC_DYLD_INFO=0x22,
                         LC_DYLD_INFO_ONLY=(0x22 | LC_REQ_DYLD),
                         LC_LOAD_UPWARD_DYLIB=(0x23 | LC_REQ_DYLD),
                         LC_VERSION_MIN_MACOSX=0x24,
                         LC_VERSION_MIN_IPHONEOS=0x25,
                         LC_FUNCTION_STARTS=0x26,
                         LC_DYLD_ENVIRONMENT=0x27,
                         LC_MAIN=(0x28 | LC_REQ_DYLD),
                         LC_DATA_IN_CODE=0x29,
                         LC_SOURCE_VERSION=0x2A,
                         LC_DYLIB_CODE_SIGN_DRS=0x2B,
                         LC_ENCRYPTION_INFO_64=0x2C,
                         LC_LINKER_OPTION=0x2D,
                         LC_LINKER_OPTIMIZATION_HINT=0x2E,
                         LC_VERSION_MIN_TVOS=0x2F,
                         LC_VERSION_MIN_WATCHOS=0x30,
                         LC_NOTE=0x31,
                         LC_BUILD_VERSION=0x32,
                         LC_DYLD_EXPORTS_TRIE=(0x33 | LC_REQ_DYLD),
                         LC_DYLD_CHAINED_FIXUPS=(0x34 | LC_REQ_DYLD),
                         LC_FILESET_ENTRY=(0x35 | LC_REQ_DYLD)
                         )


def __lc_str_from_load_command(this):
    load_command = this._._._
    return Struct(
        # this struct appears as a union in apple's opensource, but is represented in an awful manner, whereas
        # they would treat both the `ptr` and `offset` members as offsets (and never as pointers),
        # so we changed this a bit into the struct that it should have been
        'offset' / Hex(Int32ul),
        '_pad' / Int32ul,
        'name' / Pointer(load_command._start + this.offset, PaddedString(load_command.cmdsize - this.offset, 'utf8')),
    )


__dylib_t = Struct(
    'lc_str' / __lc_str_from_load_command(this),
    'timestamp' / Int64ul,
    'current_version' / Int64ul,
    'compatibility_version' / Int64ul,
)

# Load commands:
# reference - https://opensource.apple.com/source/xnu/xnu-2050.18.24/EXTERNAL_HEADERS/mach-o/loader.h
__dylib_command_t = Struct(
    'dylib' / __dylib_t,
)

__segment_command_t = Struct(
    'segname' / PaddedString(16, 'utf8'),
    'vmaddr' / SymbolFormatField(lldb.hilda_client),
    'vmsize' / Int64ul,
    'fileoff' / Int64ul,
    'filesize' / Int64ul,
    'maxprot' / Int32ul,
    'initprot' / Int32ul,
    'nsects' / Int32ul,
    'flags' / Int32ul,
)

__uuid_command_t = Struct(
    'uuid' / Array(16, Int8ul)
)

__build_tool_version = Struct(
    'tool' / Int32ul,
    'version' / Int32ul
)

__build_version_command_t = Struct(
    'platform' / Int32ul,
    'minos' / version_t,
    'sdk' / version_t,
    'ntools' / Int32ul,
    # '_build_tools' / SymbolFormatField(this),
    # 'build_tools' / If(this._build_tools != 0, Array(this.ntools, Pointer(this._build_tools, __build_tool_version)))
)

load_command_t = Struct(
    '_start' / Tell,
    'cmd' / LOAD_COMMAND_TYPE,
    'cmdsize' / Int32ul,
    '_data_offset' / Tell,
    'data' / Switch(this.cmd, {
        LOAD_COMMAND_TYPE.LC_BUILD_VERSION: __build_version_command_t,
        LOAD_COMMAND_TYPE.LC_UUID: __uuid_command_t,
        LOAD_COMMAND_TYPE.LC_SEGMENT_64: __segment_command_t,
        LOAD_COMMAND_TYPE.LC_LOAD_DYLIB: __dylib_command_t,
        LOAD_COMMAND_TYPE.LC_LOAD_WEAK_DYLIB: __dylib_command_t,
        LOAD_COMMAND_TYPE.LC_ID_DYLIB: __dylib_command_t,
        LOAD_COMMAND_TYPE.LC_REEXPORT_DYLIB: __dylib_command_t,
        LOAD_COMMAND_TYPE.LC_LAZY_LOAD_DYLIB: __dylib_command_t,
    }, Bytes(this.cmdsize - (this._data_offset - this._start))),
    Seek(this._start + this.cmdsize),
)


class LoadCommand(object):
    def __init__(self, load_command_data):
        self.__load_command_data = load_command_data
        self.__cmd = load_command_data.cmd
        self.__cmd_size = load_command_data.cmdsize
        self.__data_offset = load_command_data._data_offset

    @property
    def cmd(self):
        return self.__cmd

    @property
    def cmd_size(self):
        return self.__cmd_size


class DylibCommand(LoadCommand):
    def __init__(self, load_command_data):
        super(DylibCommand, self).__init__(load_command_data)
        dylib_data = load_command_data.data.dylib

        self.__path = dylib_data.lc_str.name
        self.__timestamp = dylib_data.timestamp
        self.__current_version = dylib_data.current_version
        self.__compatibility_version = dylib_data.compatibility_version

    def __str__(self):
        return f'<DylibCommand: path = {self.path}>'

    def __repr__(self):
        return self.__str__()

    @property
    def path(self):
        return self.__path


class Segment64Command(LoadCommand):
    def __init__(self, load_command_data):
        super(Segment64Command, self).__init__(load_command_data)

        self.__segname = load_command_data.data.segname
        self.__vmaddr = load_command_data.data.vmaddr
        self.__vmsize = load_command_data.data.vmsize
        self.__fileoff = load_command_data.data.fileoff
        self.__filesize = load_command_data.data.filesize
        self.__maxprot = load_command_data.data.maxprot
        self.__initprot = load_command_data.data.initprot
        self.__nsects = load_command_data.data.nsects
        self.__flags = load_command_data.data.flags

    @property
    def vmaddr(self):
        return self.__vmaddr

    @property
    def segname(self):
        return self.__segname

    def __str__(self):
        return f'<Segment64Command: segname={self.segname}, vmaddr={self.vmaddr}, vmsize={self.__vmsize}>'

    def __repr__(self):
        return self.__str__()


class UUIDCommand(LoadCommand):
    def __init__(self, load_command_data):
        super(UUIDCommand, self).__init__(load_command_data)

        self.__uuid = load_command_data.data.uuid

    def __str__(self):
        return '<UUIDCommand>'

    def __repr__(self):
        return self.__str__()


class BuildVersionCommand(LoadCommand):
    def __init__(self, load_command_data):
        super(BuildVersionCommand, self).__init__(load_command_data)

        self.__platform = load_command_data.data.platform
        self.__minos = load_command_data.data.platform
        self.__sdk = load_command_data.data.platform
        self.__ntools = load_command_data.data.platform
        self.__build_tools = load_command_data.data.build_tools

    def __str__(self):
        return f'<BuildVersionCommand platform={self.platform}>'

    def __repr__(self):
        return self.__str__()

    @property
    def platform(self):
        return self.__platform

    @property
    def minos(self):
        return self.__minos

    @property
    def sdk(self):
        return self.__sdk


class UnimplementedCommand(LoadCommand):
    def __init__(self, load_command_data):
        super(UnimplementedCommand, self).__init__(load_command_data)

        self.__bytes = load_command_data.data

    def __str__(self):
        return f'<cmd={self.cmd} is not implemented((Feel free to add!!). bytes={self.__bytes}>'

    def __repr__(self):
        return self.__str__()


class LoadCommands(object):
    def __init__(self, load_commands_data):
        self.__load_commands = []
        for load_command_data in load_commands_data:
            if load_command_data.cmd == 'LC_SEGMENT_64':
                load_command = Segment64Command(load_command_data)
            elif load_command_data.cmd == 'LC_LOAD_DYLIB' or \
                    load_command_data.cmd == 'LC_ID_DYLIB' or \
                    load_command_data.cmd == 'LC_LOAD_WEAK_DYLIB' or \
                    load_command_data.cmd == 'LC_REEXPORT_DYLIB' or \
                    load_command_data.cmd == 'LC_LAZY_LOAD_DYLIB':
                load_command = DylibCommand(load_command_data)
            elif load_command_data.cmd == 'LC_UUID':
                load_command = UUIDCommand(load_command_data)
            else:
                load_command = UnimplementedCommand(load_command_data)

            self.__load_commands.append(load_command)

    @property
    def all(self):
        return self.__load_commands

    @property
    def segment_commands(self):
        return [segment_command for segment_command in self.__load_commands if
                type(segment_command) == Segment64Command]

    @property
    def dylib_commands(self):
        return [dylib_command for dylib_command in self.__load_commands if type(dylib_command) == DylibCommand]

    def find(self, predicate=None):
        if predicate is None:
            return self.__load_commands

        matching_load_commands = []
        for load_command in self.load_commands:
            if predicate(load_command):
                matching_load_commands.append(load_command)

        return matching_load_commands



import logging
import os
import time
from pathlib import Path

import click
import coloredlogs

coloredlogs.install(level=logging.DEBUG)


def disable_logs():
    logging.getLogger('asyncio').disabled = True
    logging.getLogger('parso.cache').disabled = True
    logging.getLogger('parso.cache.pickle').disabled = True
    logging.getLogger('parso.python.diff').disabled = True
    logging.getLogger('humanfriendly.prompts').disabled = True
    logging.getLogger('blib2to3.pgen2.driver').disabled = True


def tunnel_local_port(port):
    execute(f'python3 -m pymobiledevice3 usbmux forward {port} {port} -d')


def execute(cmd):
    logging.debug(f'executing: {cmd}')
    return os.system(cmd)


@click.group()
def cli():
    pass


def start_remote(debug_port: int, ssh_port: int, process: str, hostname: str, rc_script: str):
    tunnel_local_port(debug_port)
    assert not execute(
        f'ssh -i ~/.ssh/id_rsa -p {ssh_port} root@{hostname} "debugserver 0.0.0.0:{debug_port} --attach={process}"&')

    # wait for it to load
    time.sleep(1)

    # connect local LLDB client
    commands = [f'process connect connect://{hostname}:{debug_port}',
                f'command script import {rc_script}']
    commands = '\n'.join(commands)
    execute(f'lldb --one-line "{commands}"')


@cli.command('remote')
@click.argument('process')
@click.argument('ssh_port', type=click.INT)
@click.option('--debug-port', type=click.INT, default=1234)
@click.option('--hostname', default='localhost')
def remote(process, ssh_port, debug_port, hostname):
    """ Start debugserver at remote device and connect using lldb """
    start_remote(debug_port, ssh_port, process, hostname,
                 os.path.join(Path(__file__).resolve().parent, "lldb_entrypoint.py"))


@cli.command('bare')
def bare():
    """ Just start an lldb shell """
    # connect local LLDB client
    commands = [f'command script import {os.path.join(Path(__file__).resolve().parent, "lldb_entrypoint.py")}']
    commands = '\n'.join(commands)
    execute(f'lldb --one-line "{commands}"')


if __name__ == '__main__':
    disable_logs()
    cli()



from construct import Struct, Int64ul, Int32ul, Int8ul, Array

uuid_t = Struct(
    'time_low' / Int64ul,
    'time_mid' / Int32ul,
    'time_hi_and_version' / Int32ul,
    'clock_seq_hi_and_reserved' / Int8ul,
    'clock_seq_low' / Int8ul,
    'node' / Array(6, Int8ul)
)



#!/usr/bin/xcrun python3

import logging

import lldb
import pytest

logging.getLogger('parso.cache').disabled = True
logging.getLogger('parso.cache.pickle').disabled = True
logging.getLogger('parso.python.diff').disabled = True
logging.getLogger('humanfriendly.prompts').disabled = True

lldb.hilda_client = None


class AddDebuggerPlugin:
    def __init__(self, debugger):
        self.__name__ = 'AddDebuggerPlugin'
        self.debugger = debugger

    @pytest.fixture(scope='session')
    def lldb_debugger(self, request):
        return request.config.pluginmanager.get_plugin('AddDebuggerPlugin').debugger


def run_tests(debugger, command, result, internal_dict):
    pytest.main(command.split(), plugins=[AddDebuggerPlugin(debugger)])


def __lldb_init_module(debugger, internal_dict):
    debugger.SetAsync(True)
    debugger.HandleCommand('command script add -f lldb_entrypoint.run_tests run_tests')
    print('Use "run_tests" command')



import os
from pathlib import Path

import click
from hilda import launch_lldb


@click.command()
@click.argument('process')
@click.argument('ssh_port', type=click.INT)
@click.option('--debug-port', type=click.INT, default=1234)
@click.option('--hostname', default='localhost')
def main(process, ssh_port, debug_port, hostname):
    """ Start debugserver at remote device and connect using lldb """
    launch_lldb.start_remote(debug_port, ssh_port, process, hostname,
                             os.path.join(Path(__file__).resolve().parent, "lldb_entrypoint.py"))


if __name__ == '__main__':
    launch_lldb.remote()
